##
## @file buildOneStageData.R
## @brief Example code to create features from QIDS and
##        average CGI
## @author Anonymized
## @note Anonymized
##

## Reverse and artificially left censor data
aLC = function (qtau, v, w=NULL){
    if (all(is.na(v))){
        return (list("v"=v, "w"=w));
    }
    if (!is.null(w)){
        ind = setdiff(1:length(v),
            intersect (which(!is.na(v)), which(!is.na(w))));
        v[ind] = NA;
        w[ind] = NA;
    }
    
    if (all(is.na(v))){
        return (list("v"=v, "w"=w));
    }
    vMax = max (v, na.rm=T);
    v = v - vMax + qtau;
    if (!is.null (w)){
        w[which(v < 0)] = NA;
    }
    v[which(v < 0)] = NA;

    if (all (is.na (v))){
        if (is.null(w)){
            return (list ("v"=v, "w"=w));
        }
        return (list("v"=v, "w"=rep(NA, length(v))));
    }
    
    # right align
    nv = length (v);
    vTmp = rep (NA, nv);
    vTmp[(nv + 1 - sum(!is.na(v))):nv] = v[which(!is.na(v))];
    v = vTmp;
    if (!is.null (w)){
        indV = which (!is.na(v));
        indW = which (!is.na(w));
        if (length(indV) != length(indW)){
            print (v);
            print (w);
        }
        wTmp = rep (NA, nv);
        wTmp[indV] = w[indW];
        w = wTmp;
    }
    return (list("v"=v, "w"=w));
}


## Setup stage one data
xtractStageOne = function (tau=0.80){
    #stard = read.csv ("star_d.csv", header=T);
    stard = read.csv ("stard_s1.csv", header=T);
    ## add average cgi, mean QIDs (outcome), and nbr comorbidities 
    n = dim (stard)[1];
    cgiNms = names(stard)[grep("CGI_I_l1_cc_[0-9]+$", names(stard))];
    yNms = names(stard)[grep("QSTOT_l2_qs_[0-9]+$", names(stard))]
    yDys = names(stard)[grep("DATE_l1_qc[0-9]+$", names(stard))];
    cLInd = which (names(stard) == "ALCOH_enr_phx");
    cUInd = which (names(stard) == "AXI_OTH_enr_phx");
    cgi = rep (NA, n);
    y1 = rep (NA, n);
    cmbd = rep (NA, n);
    for (i in 1:n){
        if (!all(is.na(stard[i,cgiNms]))){
            cgi[i] = mean (as.numeric (stard[i, cgiNms]), na.rm=T);
        }
        if (!all (is.na(stard[i,yNms]))){
            y1[i] = mean (as.numeric (stard[i, yNms]), na.rm=T);
        }
        if (!all(is.na (stard[i, cLInd:cUInd]))){
            cmbd[i] = sum(stard[i, cLInd:cUInd] >= 1, na.rm=T);
        }
    }
    stard$cgi = cgi;
    stard$y1 = y1;
    stard$cmbd = cmbd;
    
    ## Compute observation ranges for level 1
    dts = names(stard)[grep("DATE_l1_qc_[0-9]+$", names(stard))];
    lqRange = c();
    for (i in 1:n){
        iDays = as.numeric (stard[i, dts]);
        if (!all(is.na(iDays))){
            lqRange = c(lqRange, max(iDays, na.rm=T) - min (iDays,na.rm=T));
        }
    }
    qtau = 98;  # 14 weeks
    
    days = c();
    for (i in 1:n){
        days = c ( aLC (qtau, as.numeric (stard[i, dts]))$v, days);
    }
    days = days[!is.na(days)];
    days = days[days >= 0];
    cutPts = seq(0, qtau, by=2);
    cutPts = c (-Inf, cutPts);
    nCut = length (cutPts);
    W = matrix (NA, nrow=n, ncol=length(cutPts));
    l1Q = names(stard)[grep("QSTOT_l1_qs_[0-9]+$", names(stard))];
    stard$l1MnQ = rep (NA, n); # mean QIDS
    stard$l1LstQ = rep (NA, n); # last QIDS
    stard$dayMax = rep (NA, n);
    for (i in 1:n){
        rawQIDS = as.numeric (stard[i, l1Q]);
        if (any (!is.na(rawQIDS))){
            stard$l1MnQ[i] = mean (rawQIDS, na.rm=T);
            stard$l1LstQ[i] = rawQIDS[max(which(!is.na(rawQIDS)))];
            stard$dayMax[i] = max (stard[i, dts], na.rm=T);
        }
        op = aLC (qtau, as.numeric (stard[i, dts]),
            as.numeric (stard[i, l1Q]));
        iDays = op$v;
        iQIDS = op$w;
        if (any(is.na(iDays))){
            iQIDS[which(is.na(iDays))] = NA;
        }
        notNA = intersect (which(!is.na(iDays)),
            which(!is.na(iQIDS)));
        ind = as.numeric (cut (iDays[notNA], cutPts));
        W[i, ind] = iQIDS[notNA];
    }
    obs = which ((apply (!is.na (W), 1, sum) > 5) & 
        !is.na (cgi));
    times = seq (0, qtau, by=2)[which(apply(!is.na(W), 2, sum) > 10)];
    W = W[obs, which(apply(!is.na(W), 2, sum) > 10)];  # fnl predictors
    stard = stard[obs,];
    require ("refund");
    fpcaFit = fpca.sc (Y=W, var=TRUE, makePD=TRUE);
    lenT = dim(W)[2];
    ldings = fpcaFit$scores/sqrt(lenT);
    return (list ("QIDSLDNGS"=ldings, "W"=W, "stard"=stard, "times"=times));
}


getXWAY = function (tau=0.80){
    op = xtractStageOne (tau=tau);
    stard = op$stard;
    W = op$W;  # sparse representation of level 1 QIDS
    Y = 27-stard$y1;
    ## For X we'll use age, age of onset, baseline cgi
    ## age : age_enr_el
    ## age onset : dage_enr_phx
    ## mean cgi : cgi
    ## nbr comorbidities : cmbd
    ## major depressive episodes per year : epino_enr_phx/duration
    ##       where duration = age_enr_el - dage_enr_phx
    stard$duration = stard$age_enr_el - stard$dage_enr_phx;
    stard$epino_enr_phx[stard$epino_enr_phx < 0] = NA;
    X = cbind (1, stard$cgi,
        stard$cmbd, log(stard$epino_enr_phx/stard$duration),
               stard$DEGREE_enr_dm >= 3, stard$SEX_enr_el == 1);
    A = rep(NA, length (Y));
    A[which (stard$Txt_l2_ivrc==2)] = 1;
    A[which (stard$Txt_l2_ivrc==3)] = -1;
    ind = which (!is.na (Y) &
        apply(!is.na(XRed), 1, all) &
        !is.na(A));
    return (list("X"=XRed[ind,], "W"=W[ind,], "A"=A[ind], "Y"=Y[ind],
                 "aux"=cbind(stard$l1MnQ[ind], stard$l1LstQ[ind]),
                 "times"=op$times, "dayMax"=stard$dayMax[ind]));
}



## Generate plots and results for fnl proposoal using first stage
## of STAR*D
runSTARDEvals = function (){
    set.seed (1);
    source ("../../code/fnlQLearn.R");
    strd = getXWAY ();
    folds = 50;
    valFnl = numeric (folds);
    valQ = numeric (folds);
    n = length (strd$Y); 
    shuffled = sample (1:n, replace=F);
    for (k in 1:folds){
        indL = floor((k-1)*n/folds) + 1;
        indU = floor(k*n/folds);
        tst = shuffled[indL:indU];
        trn = shuffled[setdiff (1:n, tst)];

        ## fnl q-learning
        fnlQFitted = denseQLearn (strd$X[trn,], strd$W[trn,], strd$A[trn],
            strd$Y[trn]);
        piHatFnl = fnlQFitted$piHat (strd$X[tst,], strd$W[tst,]);
        pos = mean(strd$A==1);
        prop = pos*(piHatFnl > 0) + (1-pos)*(piHatFnl < 0);
        valFnl[k] = mean(strd$Y[tst]*(strd$A[tst]==piHatFnl)/prop); 

        ## std q-learning
        XQ = cbind(strd$X, strd$aux[,2]);
        p = dim(XQ)[2];
        DMat = cbind (XQ, strd$A*XQ);
        bHat = as.numeric (lm.fit (DMat[trn,], strd$Y[trn])$coefficients);
        piHatQ = sign (XQ[tst,]%*%bHat[(p+1):(2*p)]);
        propq = pos*(piHatQ > 0) + (1-pos)*(piHatQ < 0);
        valQ[k] = mean(strd$Y[tst]*(strd$A[tst]==piHatQ)/propq); 
        print (c(valFnl[k], valQ[k]));
    }
    fnlQFitted = denseQLearn (strd$X, strd$W, strd$A, strd$Y);

    ## Plot estimated coefficient function
    #pdf ("all.pdf");
    #par (mfrow=c(1,3));
    pdf ("bHat.pdf");
    par(mar=c(5, 5, 4, 2)+0.1);
    plot (strd$times,
          fnlQFitted$coefFn, type='l', lwd=3,
          xlab="time (days)",
          ylab=expression(paste(hat(gamma), "(t)")));
    dev.off ();

    ## Plot sample of raw trajectories
    pdf("QIDSRaw.pdf");
    nSubj = 50;
    fstSubj = 55;  ## Old 95
    scdSubj = 61;  ## Old 93
    #subj = 1:nSubj
    subj = (nSubj+1):(2*nSubj);
    plot (0, type='n', xlim=c (min(strd$times), max(strd$times)),
          ylim=c(min(strd$W, na.rm=T), max(strd$W, na.rm=T)),
          xlab="time (days)", ylab="QIDS"); 
    for (k in 1:nSubj){
        ind = which(!is.na(strd$W[subj[k], ]));
        points (strd$times[ind], strd$W[subj[k],ind], type='l', col='gray');
    }
    ind = which(!is.na(strd$W[fstSubj, ]));
    points (strd$times[ind], strd$W[fstSubj,ind], type='l', col='black',
            lwd=3, lty=1);
    ind = which(!is.na(strd$W[scdSubj, ]));
    points (strd$times[ind], strd$W[scdSubj,ind], type='l', col='black',
            lwd=3, lty=2);
    dev.off ();

    
    ## Plot sample of raw and pre-alignment trajectories
    pdf("QIDSRawPreAlign.pdf");
    nSubj = 50;
    #subj = 1:nSubj
    subj = (nSubj+1):(2*nSubj);
    plot (0, type='n', xlim=c (0, max(strd$dayMax)),
          ylim=c(min(strd$W, na.rm=T), max(strd$W, na.rm=T)),
          xlab="Time (days)", ylab="QIDS"); 
    for (k in 1:nSubj){
        ind = which(!is.na(strd$W[subj[k], ]));
        points (strd$times[ind] - 98 + strd$dayMax[subj[k]],
                strd$W[subj[k],ind], type='l', col='gray');
    }
    ind = which(!is.na(strd$W[fstSubj, ]));
    points (strd$times[ind] - 98 + strd$dayMax[fstSubj],
            strd$W[fstSubj,ind], type='l', col='black',
            lwd=3, lty=1);
    ind = which(!is.na(strd$W[scdSubj, ]));
    points (strd$times[ind] - 98 + strd$dayMax[scdSubj],
            strd$W[scdSubj,ind], type='l', col='black',
            lwd=3, lty=2);
    dev.off ();
    
    
    ## Plot sample of smoothed trajectories
    pdf("QIDSmooth.pdf");
    nSubj = 50;
    #subj = 1:nSubj
    nT = length(strd$times);
    plot (0, type='n', xlim=c (min(strd$times), max(strd$times)),
          ylim=c(min(strd$W, na.rm=T), max(strd$W, na.rm=T)),
          xlab="time (days)", ylab="QIDS");
    WLD = fnlQFitted$getLdings (strd$W);
    Phi = fnlQFitted$Phi
    for (k in 1:nSubj){
        ind = which(!is.na(strd$W[subj[k], ]));
        points (strd$times,
                fnlQFitted$mu +
                as.vector(sqrt(nT)*Phi%*%as.vector(WLD[subj[k],])),
                type='l', col='gray');
    }
    ind = which(!is.na(strd$W[fstSubj, ]));
    points (strd$times,
            fnlQFitted$mu +
            as.vector(sqrt(nT)*Phi%*%as.vector(WLD[fstSubj,])),
            type='l', col='black', lwd=3);
    ind = which(!is.na(strd$W[scdSubj, ]));
    points (strd$times,
            fnlQFitted$mu +
            as.vector(sqrt(nT)*Phi%*%as.vector(WLD[scdSubj,])),
            type='l', col='black', lwd=3, lty=2);
    dev.off ();
    
    shouldBootstrap = T;
    if (shouldBootstrap){
       nBoot = 1000;
       numAlpha = length (fnlQFitted$alphaHat);
       alphaMatrix = matrix (NA, nrow=nBoot, ncol=numAlpha);
       numGamma = length (fnlQFitted$gammaHat);
       gammaMatrix = matrix (NA, nrow=nBoot, ncol=numGamma);
       b = 1;
       errFlag = F;
       while (b <= nBoot){
       	   bootInd = sample (1:n, replace=T);
	   op = tryCatch({
       	      fnlQFittedBoot = denseQLearn (strd$X[bootInd,], 
	      		     strd$W[bootInd,],  strd$A[bootInd], 
			     strd$Y[bootInd]);
	      alphaMatrix[b,] = as.numeric (fnlQFittedBoot$alphaHat);
	      gammaMatrix[b,] = as.numeric (fnlQFittedBoot$gammaHat);	
	      print (paste("Bootstrap iteration: ", b));
	      b = b + 1;
	   }, warning = function (locW){
	      print (locW);
	   },
	   error = function (locE){
	      NULL;
	   },
	   finally = {
	      NULL;
           });
       }  
    }   
    browser ();
}



